package org.argouml.uml.diagram.collaboration;

import java.beans.PropertyChangeEvent;
import java.beans.VetoableChangeListener;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.Logger;
import org.argouml.model.Model;
import org.argouml.uml.CommentEdge;
import org.argouml.uml.diagram.UMLMutableGraphSupport;

public class CollabDiagramGraphModel extends UMLMutableGraphSupport
    implements VetoableChangeListener {
    
    private static final Logger LOG =
        Logger.getLogger(CollabDiagramGraphModel.class);

    public void setCollaboration(Object collaboration) {
        try {
            if (collaboration == null) {
                throw new IllegalArgumentException(
                    "A null collaboration was supplied");
            }
            if (!(Model.getFacade().isACollaboration(collaboration))) {
                throw new IllegalArgumentException(
                    "Expected a collaboration. The type received was "
                    + collaboration.getClass().getName());
            }
        } catch (IllegalArgumentException e) {
            LOG.error("Illegal Argument to setCollaboration", e);
            throw e;
        }
        setHomeModel(collaboration);
    }

    public List getPorts(Object nodeOrEdge) {
	if (Model.getFacade().isAClassifierRole(nodeOrEdge)) {
	    List result = new ArrayList();
	    result.add(nodeOrEdge);
	    return result;
	}
	return Collections.EMPTY_LIST;
    }

    
    public Object getOwner(Object port) {
	return port;
    }

    public List getInEdges(Object port) {

	if (Model.getFacade().isAClassifierRole(port)) {
	    Object cr = port;
	    Collection ends = Model.getFacade().getAssociationEnds(cr);
	    if (ends == null) {
                return Collections.EMPTY_LIST;
            }
	    List result = new ArrayList();
	    for (Object end : ends) {
		result.add(Model.getFacade().getAssociation(end));
	    }
	}
	return Collections.EMPTY_LIST;
    }
    
    public List getOutEdges(Object port) {
	return Collections.EMPTY_LIST; 
    }

    @Override
    public boolean canAddNode(Object node) {
        if (node == null) {
            return false;
        }
        if (Model.getFacade().isAAssociation(node)
                && !Model.getFacade().isANaryAssociation(node)) {
            
            return false;
        }
    
	if (containsNode(node)) {
            return false;
        }
	return (Model.getFacade().isAClassifierRole(node)
            || Model.getFacade().isAMessage(node)
            || Model.getFacade().isAComment(node));
    }

    @Override
    public boolean canAddEdge(Object edge)  {
	if (edge == null) {
            return false;
        }
	if (containsEdge(edge)) {
            return false;
        }
	Object end0 = null;
        Object end1 = null;
	if (Model.getFacade().isAAssociationRole(edge)) {
	    Collection conns = Model.getFacade().getConnections(edge);
            Iterator iter = conns.iterator();
	    if (conns.size() < 2) {
                return false;
            }
	    Object associationEndRole0 = iter.next();
	    Object associationEndRole1 = iter.next();
	    if (associationEndRole0 == null || associationEndRole1 == null) {
	        return false;
	    }
	    end0 = Model.getFacade().getType(associationEndRole0);
	    end1 = Model.getFacade().getType(associationEndRole1);
	} else if (Model.getFacade().isAGeneralization(edge)) {
	    Object gen =  edge;
	    end0 = Model.getFacade().getGeneral(gen);
	    end1 = Model.getFacade().getSpecific(gen);
	} else if (Model.getFacade().isADependency(edge)) {
	    Collection clients = Model.getFacade().getClients(edge);
	    Collection suppliers = Model.getFacade().getSuppliers(edge);
	    if (clients == null || clients.isEmpty() 
	            || suppliers == null || suppliers.isEmpty()) {
                return false;
            }
	    end0 = clients.iterator().next();
	    end1 = suppliers.iterator().next();
	} else if (edge instanceof CommentEdge) {
	    end0 = ((CommentEdge) edge).getSource();
	    end1 = ((CommentEdge) edge).getDestination();
	} else {
	    return false;       
        }
        
        if (end0 == null || end1 == null) {
            LOG.error("Edge rejected. Its ends are not attached to anything");
            return false;
        }
        
        if (!containsNode(end0)
                && !containsEdge(end0)) {
            LOG.error("Edge rejected. Its source end is attached to " + end0
                    + " but this is not in the graph model");
            return false;
        }
        if (!containsNode(end1)
                && !containsEdge(end1)) {
            LOG.error("Edge rejected. Its destination end is attached to "
                    + end1 + " but this is not in the graph model");
            return false;
        }
        
        return true;
    }

    @Override
    public void addNode(Object node) {
	LOG.debug("adding MClassifierRole node!!");
	if (!canAddNode(node)) {
            return;
        }
	getNodes().add(node);
	
	if (Model.getFacade().isAClassifier(node)) {
	    Model.getCoreHelper().addOwnedElement(getHomeModel(), node);
	    
	}

	fireNodeAdded(node);
    }

    @Override
    public void addEdge(Object edge) {
        LOG.debug("adding class edge!!!!!!");
        if (!canAddEdge(edge)) {
            return;
        }
        getEdges().add(edge);
        
        if (Model.getFacade().isAModelElement(edge)
	    && Model.getFacade().getNamespace(edge) == null) {
            Model.getCoreHelper().addOwnedElement(getHomeModel(), edge);
        }
        fireEdgeAdded(edge);
    }

    @Override
    public void addNodeRelatedEdges(Object node) {
        super.addNodeRelatedEdges(node);

	if (Model.getFacade().isAClassifier(node)) {
	    Collection ends = Model.getFacade().getAssociationEnds(node);
	    for (Object end : ends) {
		if (canAddEdge(Model.getFacade().getAssociation(end))) {
                    addEdge(Model.getFacade().getAssociation(end));
                }
	    }
	}
	if (Model.getFacade().isAGeneralizableElement(node)) {
	    Collection generalizations = 
	        Model.getFacade().getGeneralizations(node);
	    for (Object generalization : generalizations) {
		if (canAddEdge(generalization)) {
		    addEdge(generalization);
		    return;
		}
	    }
	    Collection specializations = Model.getFacade().getSpecializations(node);
	    for (Object specialization : specializations) {
		if (canAddEdge(specialization)) {
		    addEdge(specialization);
		    return;
		}
	    }
	}
	if (Model.getFacade().isAModelElement(node)) {
	    Collection dependencies =
		new ArrayList(Model.getFacade().getClientDependencies(node));
	    dependencies.addAll(Model.getFacade().getSupplierDependencies(node));
	    for (Object dependency : dependencies) {
		if (canAddEdge(dependency)) {
		    addEdge(dependency);
		    return;
		}
	    }
	}
    }
    
    @Override
    public boolean canConnect(Object fromP, Object toP) {
	if ((Model.getFacade().isAClassifierRole(fromP))
	    && (Model.getFacade().isAClassifierRole(toP))) {
            return true;
        }
	return false;
    }

    public void vetoableChange(PropertyChangeEvent pce) {
	
	if ("ownedElement".equals(pce.getPropertyName())) {
	    List oldOwned = (List) pce.getOldValue();
	    Object eo =  pce.getNewValue();
	    Object me = Model.getFacade().getModelElement(eo);
	    if (oldOwned.contains(eo)) {
		LOG.debug("model removed " + me);
		if (Model.getFacade().isAClassifier(me)) {
                    removeNode(me);
                }
		if (Model.getFacade().isAMessage(me)) {
                    removeNode(me);
                }
		if (Model.getFacade().isAAssociation(me)) {
                    removeEdge(me);
                }
	    } else {
		LOG.debug("model added " + me);
	    }
	}
    }

    private static final long serialVersionUID = -4895696235473642985L;
} 
